# Copyright (c) 2003-2005 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id: database.py,v 1.19 2005/02/01 14:10:29 zagy Exp $
"""CMF link checker tool - link database
"""


# Python imports
import socket
from md5 import md5
from xmlrpclib import ServerProxy, Fault

# Zope imports
from Products.BTreeFolder2.BTreeFolder2 import BTreeFolder2
from AccessControl import getSecurityManager, ClassSecurityInfo
from Globals import InitializeClass
import zLOG

# CMF/Plone imports
from Products.CMFCore import CMFCorePermissions

# CMFLinkChecker imports
from Products.CMFLinkChecker.interfaces import ILinkDatabase
from Products.CMFLinkChecker.LinkInfo import LinkInfo
from Products.CMFLinkChecker import permissions


WEBSERVICE_STATEMAP = {"unknown" : "grey",
                    "ok" : "green",
                    "temporary unavailable": "orange",
                    "unavailable": "red"}

def manage_addLinkDatabase(container, id):
    """Add a new link database to a link manager."""
    container._setObject(id, LinkDatabase(id))

class LinkDatabase(BTreeFolder2):
    """A database of links. 
    
       Manages the physical storage of link information.

       This link database depends on the fact that there is exactly
       one non-ILinkInfo-Object here: the catalog.

       Otherwise the amount of links will be calculated wrong.
    """

    __implements__ = (ILinkDatabase, ) + BTreeFolder2.__implements__

    defaultURLPrefix = ""

    webservice = "http://lms.gocept.com/tp1"
    clientid = "change_me"

    security = ClassSecurityInfo()

    def manage_afterAdd(self, item, container):
        # The link catalog that manages links
        self.manage_addProduct['ZCatalog'].manage_addZCatalog('link_catalog', 'Link Catalog')
        self.link_catalog.addIndex('url', 'FieldIndex')
        self.link_catalog.addIndex('state', 'KeywordIndex')
        self.link_catalog.addIndex('objects', 'KeywordIndex')
        self.link_catalog.addIndex('lastupdate', 'DateIndex')
        self.link_catalog.addIndex('registered', 'FieldIndex')
       
        self.link_catalog.addColumn("lastcheck", None)
        self.link_catalog.addColumn("lastupdate", None)
        self.link_catalog.addColumn("oldstate", None)
        self.link_catalog.addColumn("objects", [])
        self.link_catalog.addColumn("reason", "")
        self.link_catalog.addColumn("state", "")
        self.link_catalog.addColumn("url", "")
        self.link_catalog.addColumn("getId", "")

    ###############
    # ILinkDatabase

    security.declareProtected(CMFCorePermissions.ManagePortal, 'configure')
    def configure(self, defaultURLPrefix=None, clientid=None, webservice=None):
        """Set configuration values."""
        if clientid is not None:
            self.clientid = clientid
        if defaultURLPrefix is not None:
            self.defaultURLPrefix = defaultURLPrefix
        if webservice is not None:
            self.webservice = webservice
       
        # need to reconfigure connection
        try:
            delattr(self, '_v_lms_client')
        except KeyError:
            pass

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'registerLink')
    def registerLink(self, link, object):
        """Registers a link/object pair at the database.

           Returns None.
        """
        links = (link, )
        self.registerManyLinks(links, object)

    security.declareProtected(permissions.USE_LINK_MANAGEMENT,
                              'registerManyLinks')
    def registerManyLinks(self, links, object):
        """Registers may link/object pair at the database.

           Returns None.
        """
        sm = getSecurityManager()
        if not sm.checkPermission(CMFCorePermissions.ModifyPortalContent,
                                  object):
            # XXX import exception
            raise "Unauthorized", \
                "Can't modify link registrations for this object."

        for link in links:
            self._register_link(link, object)
        self._updateWSRegistrations()

    def _register_link(self, link, object):
        id = self._hashLink(link)
        info = getattr(self, id, None)
        if info is None:
            info = LinkInfo(link, id)
            self._setObject(id, info)
            info = getattr(self, id)
        info.addObject(object)
        return info

    security.declarePrivate("_updateWSRegistrations")
    def _updateWSRegistrations(self):
        """Registers a link with the webservice. 
        
        May also register yet non-registered other links.
        """
        unregistered = [ x.url for x in self.query(registered=False) ]
        if not unregistered:
            # nothing to do
            return
        lms = self._getWebServiceConnection()
        if lms is None:
            # no connection to the server
            return 

        states = lms.registerManyLinks(unregistered)
        for url, state, reason in states:
            link = self.byLink(url)
            link.registered = True
            state = WEBSERVICE_STATEMAP[state]
            link.updateStatus(state, reason)

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'unregisterLink')
    def unregisterLink(self, link, object):
        """Unregisters a given link/object pair at the database.

           If the given link isn't referenced by any object anymore, it will 
           be removed from the database completely.

           Raises KeyError if the given pair isn't registered.

           Returns None.
        """
        sm = getSecurityManager()
        if not sm.checkPermission(CMFCorePermissions.ModifyPortalContent,
                                  object):
            raise "Unauthorized", \
                "Can't modify link registrations for this object."

        id = self._hashLink(link)
        try:
            info = getattr(self, id)
        except AttributeError, m:
            raise KeyError, m

        info.removeObject(object)

    def _getWebServiceConnection(self):
        """return connection to lms

        returns LMSClient instance if connection is ok
        returns None if connection is not Ok
        """
        client = self._get_connection_object()
        ok, error = client.checkConnection()
        if not ok:
            zLOG.LOG('CMFLinkChecker', zLOG.ERROR,
                     'Connection to LMS not possible: %s' % (error, ))
            client = None
        return client

    def _get_connection_object(self):
        client = getattr(self, '_v_lms_client', None)
        if client is None:
            client = LMSClient(self.webservice, self.clientid)
            self._v_lms_client = client
        return client

    def _deleteLinks(self, ids):
        # Unregister with LMS
        lms = self._getWebServiceConnection()
        if lms is None:
            return
        
        for id in ids:
            url = self[id].url
            lms.unregisterLink(url)

        self.manage_delObjects(ids)

    security.declareProtected(permissions.USE_LINK_MANAGEMENT,
                              'unregisterObject')
    def unregisterObject(self, object):
        """Unregisters all links for this object.

           Returns None.
        """
        sm = getSecurityManager()
        if not sm.checkPermission(CMFCorePermissions.ModifyPortalContent,
                                  object):
            raise "Unauthorized", \
                "Can't modify link registrations for this object."

        links = self.byObject(object)
        for link in links:
            # XXX Optimize for speed when needed
            self.unregisterLink(link.url, object)

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'byLink')
    def byLink(self, link):
        """Retrieve information about a link.

           Raises KeyError if the given link isn't registered yet.

           Returns an ILinkInfo object.
        """
        id = self._hashLink(link)
        try:
            info = getattr(self, id)
        except AttributeError, m:
            raise KeyError, m
        return info

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'byObject')
    def byObject(self, object, state=None):
        """Retrieve information about an object.

           Returns a list of ILinkInfo objects. 
           Returns an empty list if the object hasn't been registered yet.
        """
        sm = getSecurityManager()
        if not sm.checkPermission(CMFCorePermissions.AccessContentsInformation,
                                  object):
            raise "Unauthorized", "Can't view links on this object."
     
        lc = self.getLinkManager()
        uid = lc.getUIDForObject(object)

        query_args = {}
        if state is not None:
            query_args['state' ] = state

        links = self.link_catalog(objects=[uid], **query_args)
        links = [ x.getObject() for x in links ]
        return links

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'getAllLinks')
    def getAllLinks(self):
        """Returns a list of all ILinkInfo objects.

           Warning: This is likely to be a slow method in large data sets.
        """
        return self.objectValues(['Link Information'])

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'getAllLinkIds')
    def getAllLinkIds(self):
        """Returns a list of the ids of all ILinkInfo objects.

           Warning: This is likely to be a slow method in large data sets.
        """
        return self.objectIds(['Link Information'])


    security.declareProtected(permissions.USE_LINK_MANAGEMENT,
                              'getLinkIterator')
    def getLinkIterator(self):
        """returns generator for all link objects"""
        for id in self.getAllLinkIds():
            yield self.get(id)
        

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'getLinkCount')
    def getLinkCount(self):
        """Returns the amount of links currently in the database."""
        # Never,
        # Never,
        # Never put anything more than the catalog in here!
        return len(self) - 1

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'query')
    def query(self, **args):
        """Returns the result of querying the ZCatalog that indexes
           the link database.
        """
        return self.link_catalog(**args)

    security.declareProtected(permissions.USE_LINK_MANAGEMENT, 'getCatalog')
    def getCatalog(self):
        """Returns the used catalog.
        """
        return self.link_catalog

    security.declareProtected(CMFCorePermissions.ManagePortal,
        'removeAllRegistrations')
    def removeAllRegistrations(self):
        """Removes all object registrations from the known links."""
        for link in self.objectValues(["Link Information"]):
            link.registered = False
            link.clearObjects()

    security.declareProtected(CMFCorePermissions.ManagePortal, 'cleanUp')
    def cleanUp(self):
        """Removes all links without object registrations."""
        items = self.link_catalog(objects=[])
        ids = [ item.getId for item in items ]
        self._deleteLinks(ids)

    # XXX: security?
    def updateLinkStatus(self, url, state, reason):
        """XML-RPC connector for LMS"""
        state = WEBSERVICE_STATEMAP[state]
        try:
            link = self.byLink(url)
        except KeyError:
            pass
        else:
            link.updateStatus(state, reason)

    # XXX: security?
    def updateManyStates(self, update_list):
        """XML-RPC connector for LMS"""
        for url, state, reason in update_list:
            self.updateLinkStatus(url, state, reason)

    security.declareProtected(CMFCorePermissions.ManagePortal,
                             'updateAllStatus')
    def updateAllStatus(self):
        """Manually update all link status'
        """
        server = self._getWebServiceConnection()
        if server is None:
            return
        for item in self.link_catalog():
            try:
                status, reason = server.getStatus(item.url)
            except KeyError:
                # not registered with server
                server.registerLink(item.url)
                status = 'temporary unavailable'
                reason = ''
                
            status = WEBSERVICE_STATEMAP[status]
            if item.state != status:
                item = item.getObject()
                item.updateStatus(status, reason)
   
    security.declareProtected(permissions.USE_LINK_MANAGEMENT,
                              'checkConnection')
    def checkConnection(self):
        """return if connection is up and running

            returns r.ok == True if connection is ok
            returns r.ok == False if connection is *not* ok
                sets also r.error to some error information
        """
        class R:
            __allow_access_to_unprotected_subobjects__ = 1
            ok = True
            error = ''
            
        server = self._get_connection_object()
        ok, error = server.checkConnection()
        r = R()
        r.ok = ok
        r.error = error
        return r
   
    #################
    # private methods

    security.declarePrivate('_hashLink')
    def _hashLink(self, url):
        """Provides a hash which can be used as an id for a url."""
        if url is None:
            url = ""
        if isinstance(url, unicode):
            url = url.encode('utf-8')
        return md5(url).hexdigest()

InitializeClass(LinkDatabase)


class LMSClient:
    """client to lms"""

    _exception_mapping = {
        'KeyError': KeyError,
    }

    def __init__(self, url, client_id):
        self.url = url
        self.client_id = client_id
        self._server = ServerProxy(url)

    def registerLink(self, url):
        return self._server.registerLink(url, self.client_id)

    def unregisterLink(url):
        return self._server.unregisterLink(url, self.client_id)

    def registerManyLinks(self, urls):
        result = []
        if urls:
            result = self._server.registerManyLinks(urls, self.client_id)
        return result

    def getStatus(self, url):
        try:
            status = self._server.getStatus(url)
        except Fault, f:
            raise self._translate_fault(f)
        return status

    def setClientNotifications(self, status):
        self._server.setClientNotifications(self.client_id, status)


    def checkConnection(self):
        is_ok, error = False, 'unknown'

        try:
            is_ok, error = self._server.checkConnection(self.client_id)
        except Fault, f:
            is_ok, error = False, '%s (%s)' % (f.faultString, f.faultCode)
        except socket.error, e:
            e_code, e_string = e.args
            is_ok, error = False, 'socket.error: %s (%s)' % (e_string, e_code)
        except Exception, e:
            is_ok, error = False, str(e)

        return is_ok, error

    #########
    # private

    def _translate_fault(self, f):
        name, value = f.faultCode, f.faultString

        exception_class = self._exception_mapping.get(name)
        if exception_class is None:
            e = f
        else:
            e = exception_class(value)
            
        return e
        
        

        
    
