##############################################################################
#
# Copyright (c) 2003 gocept gmbh & co. kg. All rights reserved.
#
# See also LICENSE.txt
#
##############################################################################
"""CMF link checker tool interface definitions

$Id: interfaces.py,v 1.12 2005/01/26 10:03:36 zagy Exp $"""
    
from Interface import Interface, Attribute

class ILinkManager(Interface):

    database = Attribute("A link database implementing the "
                         "ILinkDatabase interface.")
    reports = Attribute("A report utility implementing the "
                        "IBaseReports interface.")
    retrieving = Attribute("A retrieve manager implementing the "
                           "IRetrieveManager interface.")

    def shortURL(url, target=50):
        """Context aware wrapper for creating a shortened URL"""

    def isManaged(object):
        """Tells if the object is managed by the link manager."""

    def resolveRelativeLink(url, context):
        """resolve realtive url from context to an absolute url

        This is simpilar to absolute_url, but does not require an object and
        honours the prefix settings from the Link Management settings
        """
       
    def isUserAllowed():
        """returns whether the user is allowed to use Link Management services

        returns True if the currently logged in user is allowed to use Link
            Management services at all, False otherwise.

        It's checked if the users has the 'Use link management functions' 
        permission in the Plone root
        """
        

    def getLinkManager():
        """return self"""

class IUIDManager(Interface):

    def getUIDForObject(object):
        """Return the UID (string) for this object.
       

        return None if object is None
        
        the UID manager will not create a UID if there
        is none.
        """

    def geTUIDForBrain(brain):
        """return the UID (string) for the brain's object
        
            return None if there is no UID
            NOTE: this will *not* create a uid
        """

    def getObjectForUID(uid):
        """Return the object corresponding to this UID.

           Raises KeyError if the UID doesn't correspond to an object.

           Returns an object.
        """

class IBaseReports(Interface):
    """Supplement functions for various basic reports."""

    def MailNotifications(user):
        """Runs the mail notification."""

    def GroupedLinksForAuthenticatedMember():
        """Returns a list of links aggregated by state for the current user.
        
           The list only features catalog brains from the database.
        """

    def DocumentsInState(state):
        """Returns a list of all objects that contain links in the given state."""

    def LinksInState(state):
        """Returns a list of links in the given state."""

    def ManagementOverview():
        """Returns a comprehensive management overview over the complete site."""


class ILinkDatabase(Interface):
    """A database of links. 
    
       Manages the physical storage of link information.
    """

    defaultURLPrefix = Attribute("A valid URL that is used (component wise) to "
                                 "prepend incomplete links")

    def configure(defaultURLPrefix=None):
        """Set configuration values."""

    def registerLink(link, object):
        """Registers a link/object pair at the database.

           Returns None.
        """

    def unregisterLink(link, object):
        """Unregisters a given link/object pair at the database.

           If the given link isn't referenced by any object anymore, it will 
           be removed from the database completely.

           Raises KeyError if the given pair isn't registered.

           Returns None.
        """

    def unregisterObject(object):
        """Unregisters all links for this object.

           Returns None.
        """

    def byLink(link):
        """Retrieve information about a link.

           Raises KeyError if the given link isn't registered yet.

           Returns an ILinkInfo object.
        """

    def byObject(object):
        """Retrieve information about an object.

           Returns a list of ILinkInfo objects. 
           Returns an empty list if the object hasn't been registered yet.
        """

    def getAllLinks():
        """Returns a list of all ILinkInfo objects.

           Warning: This is likely to be a slow method in large data sets.
        """

    def getAllLinkIds():
        """Returns a list of the ids of all ILinkInfo objects.

           Warning: This is likely to be a slow method in large data sets.
        """

    def getLinkCount():
        """Returns the amount of links currently in the database."""

    def query(**args):
        """Returns the result of querying the ZCatalog that indexes
           the link database.
        """

    def getCatalog():
        """Returns the used catalog.
        """

    def removeAllRegistrations():
        """Removes all object registrations from the known links."""

    def cleanUp():
        """Removes all links without object registrations."""


class IRetrieveManager(Interface):
    """Local registry that manages the mapping between portal_type and 
       specific retrievers for a single plone instance and allows to trigger
       retrieving for an object or the whole site.
    """

    def registerRetriever(portal_type, retriever_name):
        """Registers the retriever to be called on objects
           of portal_type.

           retriever_name must be the name of a retriever as registered at the
           global retriever registry.

           Raises KeyError if the portal_type or retriever name is invalid.

           If the portal_type already has a registered retriever, the
           registration will be overwritten.
           
           Returns None.
        """

    def isRetrieverForType(retriever_name, portal_type):
        """Tells if the given retriever name is the retriever
           for this type.

           Returns True or False.
        """

    def listSupportedTypes():
        """Returns a list of portal_types that are currently supported by
           this retrieve manager.
        """

    def retrieveObject(object):
        """Retrieves the links from the given object."""

    def retrieveSite():
        """Retrieves the links from all objects in the site."""
 
    def supportsRetrieving(object):
        """Tells if the object is supported for retrieving links."""

    def getAllRetrieverNames():
        """Returns the name of all available retrievers."""

class IGlobalRetrieverRegistry(Interface):
    """Process global registry that holds information
       about all system wide available retrievers.
    """

    def register(method, name, defaults=[]):
        """Register a retriever method globally under the given title and as a 
           default for the given portal_types.

           Defaults are overwritten subsequently by later calls.

           Raises ValueError if the method has been registered already.

           Returns None
        """

    def getByName(name):
        """Returns a retriever by it's name.

           Raises KeyError if a retriever with this name doesn't exists.
        """

    def getDefault(type):
        """Returns the default retriever for a type.

           Raises KeyError if there is no retriever for the given type.
        """

    def listNames():
        """Returns a list of all retriever names."""
        

class ILinkInformation(Interface):
    """Information about a link."""

    objects = Attribute("Contains a list of UIDs for objects containing this link.")
    url = Attribute("Contains the URL the link refers to.")
    lastcheck = Attribute("DateTime when the last check was performed.") 
    lastupdate = Attribute("DateTime when the last status change ocurred.")
    laststate = Attribute("The last state the link was in before switching to the current.")
    state = Attribute("The current state of the link. Can be one of ['red', " \
                      "'green', 'orange', 'grey'].")
    reason = Attribute("A verbose reason describing the error")
    registered = Attribute("Has this link been registered with the web service?")

    def addObject(object):
        """Add this uid to the object list."""
        
    def removeObject(object):
        """Remove this object from the object list.
        
           Raises KeyError if the object isn't registered.
        """

    def getObjects():
        """Returns a list of objects referencing this link."""

    def updateStatus(state, reason):
        """Updates the status information for this link."""

class IRetriever(Interface):
    """A class that implements a routine to find all links from a given object."""

    name = Attribute("Name of the retriever, suitable to identify it.")
    defaults = Attribute("A list of portal_types this retriever is a default for.")

    def __call__(self, object):
        """Finds all links from the object and return them.

           Returns a list of URLs.
        """
