# Copyright (c) 2003-2005 gocept gmbh & co. kg
# See also LICENSE.txt
# $Id: reports.py,v 1.11 2005/02/02 12:50:31 zagy Exp $
"""CMF link checker tool - basic reports
"""

# Python imports
import time

# Zope imports
from OFS.SimpleItem import SimpleItem
from Globals import InitializeClass
from AccessControl import ClassSecurityInfo, getSecurityManager
from DateTime import DateTime

# CMF/Plone imports
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.utils import getToolByName

# CMFLinkChecker imports
from Products.CMFLinkChecker.interfaces import IBaseReports
from Products.CMFLinkChecker import permissions

def manage_addBaseReports(container, id):
    """Add basic reports to a link manager."""
    container._setObject(id, BaseReports(id))

class BaseReports(SimpleItem):
    """Supplement functions for various basic reports."""

    __implements__ = (IBaseReports,) + SimpleItem.__implements__
    security = ClassSecurityInfo()

    security.declareProtected(permissions.RUN_MAIL_NOTIFICATIONS,
                              'MailNotifications')
    def MailNotifications(self):
        """Runs the mail notification.
        """
        hour = str(int(time.strftime("%H")))  # remove trailing 0 from the hour
        mailhost = getToolByName(self, 'MailHost')
        pms = getToolByName(self, "portal_membership")
        lc = self.getLinkManager()
        props = getToolByName(self, 'portal_properties')
        default_charset = props.site_properties.default_charset
        email_from_name = self.email_from_name
        email_from_address = self.email_from_address


        # Count the mails for admin notification
        mail_count = 0

        members = pms.listMembers()
        for member in members:
            # Check if the user can receive reports at all and if he wants a 
            # report *now*
            email = member.getProperty('email', None)
            if email is None:
                continue
            frequency = member.getProperty('lc_notify_frequency', [])
            if hour not in frequency:
                continue

            report = self.getGroupedLinksFor(member.getId())
            self._filter_report(member, report)
   
            # Now get the template and create the html for the output
            template = self.lc_report_template

            # sorting
            states = [ x for x in ['red', 'orange', 'grey', 'green']
                         if x in report ]
            
            mailbody = template(states=states, report=report, member=member,
                                hour=hour)
            member.setMemberProperties(
                                {'lc_notify_last_notification': DateTime()})

            mMsg = "To: %s\n" % email
            mMsg = mMsg + "From: %s <%s>\n" % (email_from_name,
                                               email_from_address)
            mMsg = mMsg + "Mime-Version: 1.0\n"
            mMsg = mMsg + "Content-Type: text/html; charset=%s\n\n" % (
                                                            default_charset, )
            mMsg = mMsg + mailbody
            mSubj = "Link report for %s o'clock" % hour
            mailhost.send(mMsg, subject=mSubj, encode='base64')
            mail_count += 1

        return "Sent Mails: %s" % mail_count

    def _filter_report(self, member, report):
        report_states = member.getProperty('lc_notify_details')
        for state in report:
            if state not in report_states:
                del report[state]
        
        changes_only = member.getProperty('lc_notify_changes_only', False)
        last_report = member.getProperty('lc_notify_last_notification',
                                         DateTime())
        for state, links in report.items():
            links_iterate = links[:]
            for id, link in enumerate(links_iterate):
                if changes_only and link.lastupdate < last_report:
                    # remove already reported links (did not change)
                    del links[id]
                    continue
                if changes_only and link.laststatus in report:
                    # add some extra information
                    report[link.laststatus].append(link)



    security.declarePublic('GroupedLinksForAuthenticatedMember')
    def GroupedLinksForAuthenticatedMember(self):
        """Returns a list of links aggregated by state for the current user.
        
           The list only features catalog brains from the database.
        """
        user = getSecurityManager().getUser().getId()
        return self.getGroupedLinksFor(user)
    
    security.declareProtected(CMFCorePermissions.ManagePortal,
                              'getGroupedLinksFor')
    def getGroupedLinksFor(self, user_id):

        catalog = getToolByName(self, 'portal_catalog')
        lc = self.getLinkManager()
        types = lc.retrieving.listSupportedTypes()

        # Find all objects this user is responsible for
        docs = catalog(Creator=user_id, portal_type=types)
        uids = [ lc.getUIDForBrain(x) for x in docs ]
        uids = [ uid for uid in uids if uid is not None]
        links = lc.database.query(objects=uids)

        groups = {'red': [], 'orange': [], 'grey': [], 'green': []}
        for link in links:
                groups[link.state].append(link)
                        
        return groups

        
    
    security.declareProtected(CMFCorePermissions.ManagePortal,
                              'DocumentsInState')
    def DocumentsInState(self, state):
        """Returns a list of all objects that contain links in state
        """
        lc = self.getLinkManager()
        documents = {}
        
        def _total_links(doc):
            uid = lc.getUIDForBrain(doc)
            links = lc.database.query(objects=[uid])
            return len(links)
            
        for link, doc, member in self._document_iterator(state):
            key = doc.getURL()
            item = documents.get(key)
            if item is None:
                item = documents[key] = {}
                item['document'] = doc
                item['total'] = _total_links(doc)
                item['instate'] = 0
                if member is None:
                    item["owner_mail"] = ""
                    item["owner"] = doc.Creator
                else:
                    item["owner_mail"] = member.getProperty("email")
                    item["owner"] = member.getProperty("fullname", doc.Creator)

            item["instate"] += 1

        return documents.values()

    security.declareProtected(CMFCorePermissions.ManagePortal, "LinksInState")
    def LinksInState(self, state):
        """Returns a list of links in the given state."""

        for link, doc, member in self._document_iterator(state):
            item = {}
            item["url"] = link.url
            item["reason"] = link.reason
            item["lastcheck"] = link.lastcheck
            item["document"] = doc
           
            if member is None:
                item["owner_mail"] = ""
                item["owner"] = doc.Creator
            else:
                item["owner_mail"] = member.getProperty("email")
                item["owner"] = member.getProperty("fullname", doc.Creator)
                
            yield item

    security.declareProtected(CMFCorePermissions.ManagePortal,
                              'ManagementOverview')
    def ManagementOverview(self):
        """Returns a comprehensive management overview over the complete site
        """
        return ManagementReport(self)
        
    def _document_iterator(self, state):

        member_cache = {}
        
        lc = self.getLinkManager()
        catalog = getToolByName(self, 'portal_catalog')
        pms = getToolByName(self, 'portal_membership')

        _marker = object()

        links = lc.database.query(state=[state])
        result = []
        for link in links:
            doc_uids = link.objects
            if not doc_uids:
                continue
            docs = catalog(portal_linkchecker_uid=doc_uids)
            for doc in docs:
                creator = doc.Creator
                member = member_cache.get(creator, _marker)
                if member is _marker:
                    member = pms.getMemberById(creator)
                    member_cache[creator] = member
                yield link, doc, member
            
    
InitializeClass(BaseReports)





class ManagementReport:

    __allow_access_to_unprotected_subobjects__ = 1

    states = [
        ('red', 'Red'),
        ('orange', 'Orange'),
        ('green', 'Green'),
        ('grey', 'Grey'),
    ]

    def __init__(self, context):
        self._context = context
        self._generate_data()

    def _generate_data(self):
        lc = self._context.getLinkManager()
   
        link_document_map = {}
        link_document_count = []
        link_state_map = {'red': 0, 'grey': 0, 'green': 0, 'orange': 0}
        document_state_map = {'red': {}, 'grey': {}, 'green': {}, 'orange': {}}

        
        for link in lc.database.getLinkIterator():
            link_document_count.append(len(link.objects))
            link_state_map[link.state] += 1
            
            for uid in link.objects:
                count = link_document_map.get(uid, 0)
                count += 1
                link_document_map[uid] = count
                document_state_map[link.state][uid] = True

        self.totalDistinctLinks = len(link_document_count)
        self.totalLinks = sum(link_document_count)
        self.totalDocuments = len(link_document_map)
       
        if not link_document_count:
            link_document_count = [0]

        self.documentsPerLinkMin = min(link_document_count)
        self.documentsPerLinkMax = max(link_document_count)
        self.documentsPerLinkAvg = float(self.totalDocuments) / self.totalLinks
      

        link_document_map_list = link_document_map.values()
        if not link_document_map_list:
            link_document_map_list = [0]

        self.linksPerDocumentMin = min(link_document_map_list)
        self.linksPerDocumentMax = max(link_document_map_list)
        self.linksPerDocumentAvg = float(self.totalLinks) / self.totalDocuments
        
        self.links = link_state_map
        self.linksPct = {}
        for state, count in link_state_map.items():
            self.linksPct[state] = float(count) / self.totalLinks * 100


        self.documents = {}
        self.documentsPct = {}
        for state, docs in document_state_map.items():
            count = len(docs)
            self.documents[state] = count
            self.documentsPct[state] = float(count) / self.totalDocuments * 100
        
